/*
    Copyright (c) 2012 digitalSTROM.org, Zurich, Switzerland

    Author: Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of po2json.

    po2json is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    po2json is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with po2json. If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <vector>

using namespace std;

string quotedPart(string& line)
{
    int strStart = line.find('"');
    if ((strStart == line.npos) ||
        (line[line.length() - 1] != '"') ||
        (strStart == (line.length() - 1))) {
        cerr << "quote error: " << line << endl;
        exit(1);
    }
    return line.substr(strStart + 1, line.length() - strStart - 2);
}

string splitHeader(string& line)
{
    int strStart = line.find("\"");
    if (strStart != 0) {
        cerr << "quote error: " << line << endl;
        exit(1);
    }
    int strEnd = line.find("\\n\"");
    if (strEnd == line.npos) {
        cerr << "quote error: " << line << endl;
        exit(1);
    }
    int colonPos = line.find(":");
    if (colonPos == line.npos) {
        cerr << "format error: " << line << endl;
        exit(1);
    }
    return line.substr(0, colonPos) + "\":\"" + line.substr(colonPos + 1, strEnd - colonPos - 1) + "\"";
}

typedef enum {
    eMessageId,
    eMessageIdCont,
    eMessageIdPluralCont,
    eMessageString,
    eHeaderSkip,
    eHeader,
    eFinish,
} parserState;


/*
 * {
 *      "": { ... (Header) },
 *      "Wake up":[null,"Wake Up"],
 *      "Zone":["Zones","",""]
 * }
 */

struct unit_t {
    unit_t() : context(""), id(""), plural(""), plurals(-1) {};
    string context;
    string id;
    string plural;
    vector<string> translations;
    int plurals;
};

typedef struct {
    vector<string> headers;
    vector<unit_t> units;
} translation_t;

static translation_t translation;

parserState poParser(parserState state, string& line)
{
    static struct unit_t* unit = NULL;
    if ((line.length() > 1) && (line[0] == '#')) {
        return state;
    }

    switch (state) {
    case eMessageIdCont: {
        if (line[0] == '"') {
            unit->id += quotedPart(line);
            break;
        }
    }
    case eMessageIdPluralCont: {
        if (line[0] == '"') {
            unit->plural += quotedPart(line);
            break;
        }
    }
    case eMessageId: {
        if (line == "msgid \"\"") {
            state = eHeaderSkip;
        } else if (line.find("msgctxt") == 0) {
            translation.units.push_back(unit_t());
            unit = &translation.units.back();
            unit->context = quotedPart(line);
        } else if (line.find("msgid ") == 0) {
            if (unit == NULL) {
                translation.units.push_back(unit_t());
                unit = &translation.units.back();
            }
            unit->id += quotedPart(line);
            state = eMessageIdCont;
        } else if (line.find("msgid_plural ") == 0) {
            unit->plural += quotedPart(line);
            state = eMessageIdPluralCont;
        } else if (line.find("msgstr") == 0) {
            unit->plurals++;
            unit->translations.push_back(quotedPart(line));
            state = eMessageString;
        }
        break;
    }
    case eMessageString: {
        if (line[0] == '"') {
            string& s = unit->translations.back();
            s += quotedPart(line);
            break;
        } else if (line.find("msgstr") == 0) {
            int index = strtol(line.substr(7).c_str(), NULL, 0);
            if (index != unit->plurals) {
                unit->plurals++;
                unit->translations.push_back(quotedPart(line));
            }
        } else {
            /* unit done */
            unit = NULL;
            state = eMessageId;
        }
        break;
    }
    case eHeaderSkip: {
        if (line == "msgstr \"\"") {
            state = eHeader;
        } else if (line[0] == '"') {
            /* first element of multiline msgid */
            if (unit == NULL) {
                translation.units.push_back(unit_t());
                unit = &translation.units.back();
            }
            unit->id += quotedPart(line);
            state = eMessageIdCont;
        } else {
            cerr << "expected msgstr \"\" or msgid: received instead: " << line << endl;
            exit(1);
        }
        break;
    }
    case eHeader: {
        if (line[0] == '"') {
            static string header_fragment;
            if (line.find("\\n\"") == line.npos) {
                header_fragment = line;
            } else {
                string tmp = line;
                if (!header_fragment.empty()) {
                    tmp = header_fragment.substr(0, header_fragment.length() - 1) + line.substr(1, line.length() - 1);
                    header_fragment.clear();
                }
                translation.headers.push_back(splitHeader(tmp));
            }
        } else {
            state = eMessageId;
        }
        break;
    }
    case eFinish: {
        cout << "{\"\":{";
        vector<string>::iterator it = translation.headers.begin();
        cout << *it;
        it++;
        for (; it != translation.headers.end(); ++it) {
            cout << ',' << *it;
        }
        cout << "}";
        for (vector<unit_t>::iterator it = translation.units.begin();
             it != translation.units.end();
             ++it) {
            cout << ",\"";
            if (!it->context.empty()) {
                cout << it->context << '\004';
            }
            cout << it->id << "\":[";
            if (it->plural.empty()) {
                cout << "null";
            } else {
                cout << '"' << it->plural << '"';
            }
            for (vector<string>::iterator msg = it->translations.begin();
                 msg != it->translations.end();
                 ++msg) {
                cout << ",\"" << *msg << '"';
            }
            cout << ']';
        }
        cout << "}";
    }
    }
    return state;
}

int main(int argc, char *argv[])
{
    if (argc != 2) {
        cerr << "needs .po file as first argument" << endl;
        return 1;
    }
    ifstream ifs(argv[1]);
    string temp;

    parserState state = poParser(eMessageId, temp);
    while (getline(ifs, temp)) {
        state = poParser(state, temp);
    }
    poParser(eFinish, temp);
    return 0;
}
